# [(canvas key, title, pretty title), ...]
OUTPUT_KEYS = [
    ('name', 'name', 'Name'),
    ('id', 'canvas_id', 'Canvas ID'),
    ('points_possible', 'points_possible', 'Points Possible'),
    ('published', 'published', 'Published'),
    ('due_at', 'due_at', 'Due At'),
    ('description', 'description', 'Description'),
]



import ast
import sys

import canvas.api.assignment.uploadscores
import canvas.config

DEFAULT_SKIP_ROWS = 0

def run_cli(assignment = None, path = None, skip_rows = DEFAULT_SKIP_ROWS, **kwargs):
    users, scores, comments = _load_scores(path, skip_rows)
    score_count = canvas.api.assignment.uploadscores.request(assignment = assignment,
            users = users, scores = scores, comments = comments,
            **kwargs)

    print("Uploaded %d Scores" % (score_count))
    return 0

def _load_scores(path, skip_rows):
    users = []
    scores = []
    comments = []

    with open(path, 'r') as file:
        lineno = 0
        for line in file:
            lineno += 1

            if (lineno <= skip_rows):
                continue

            line = line.strip()
            if (line == ""):
                continue

            parts = [part.strip() for part in line.split("\t")]
            if (len(parts) not in [2, 3]):
                raise ValueError(
                    "File '%s' line %d has the incorrect number of values." % (path, lineno)
                    + " Expecting 2-3, found %d." % (len(parts)))

            score = None
            if (parts[1] != ''):
                try:
                    score = ast.literal_eval(parts[1])
                except Exception as ex:
                    raise ValueError(
                        "File '%s' line %d has a score that cannot be" % (path, lineno)
                        + " converted to a number: '%s'." % (parts[1])) from ex

            users.append(parts[0])
            scores.append(score)

            if (len(parts) >= 3):
                comments.append(parts[2].strip())
            else:
                comments.append(None)

    return users, scores, comments

def main():
    config = canvas.config.get_config(exit_on_error = True, modify_parser = _modify_parser, course = True)
    return run_cli(**config)

def _modify_parser(parser):
    parser.description = 'Upload scores (and optional comments) for an assignment.'

    parser.add_argument('--skip-rows', dest = 'skip_rows',
        action = 'store', type = int, default = DEFAULT_SKIP_ROWS,
        help = 'The number of header rows to skip (default: %(default)s).')

    parser.add_argument('assignment', metavar = 'ASSIGNMENT',
        action = 'store', type = str,
        help = 'The query for the assignment to upload scores to.')

    parser.add_argument('path', metavar = 'PATH',
        action = 'store', type = str,
        help = 'Path to a TSV file where each row has 2-3 columns: email, score, and comment (optional).')

    return parser

if (__name__ == '__main__'):
    sys.exit(main())



"""
The `canvas.cli.user` package contains tools for interacting with Canvas users.
"""

import sys

import canvas.config
import canvas.util.cli

def main():
    canvas.util.cli.auto_list(default_parser = canvas.config.get_argument_parser())
    return 0

if (__name__ == '__main__'):
    sys.exit(main())



import sys

import canvas.api.assignment.uploadscores
import canvas.config

def run_cli(assignment = None, user = None, score = None, comment = None, **kwargs):
    score_count = canvas.api.assignment.uploadscores.request(
            assignment = assignment, users = [user],
            scores = [score], comments = [comment],
            **kwargs)

    print("Uploaded %d Scores" % (score_count))
    return 0

def main():
    config = canvas.config.get_config(exit_on_error = True, modify_parser = _modify_parser, course = True)
    return run_cli(**config)

def _modify_parser(parser):
    parser.description = 'Upload a single score (and optional comment) for an assignment.'

    parser.add_argument('assignment', metavar = 'ASSIGNMENT',
        action = 'store', type = str,
        help = 'The query for the assignment to upload scores for.')

    parser.add_argument('user', metavar = 'USER',
        action = 'store', type = str,
        help = 'The query for the user to upload scores for.')

    parser.add_argument('score', metavar = 'SCORE',
        action = 'store', type = float,
        help = 'Score to upload.')

    parser.add_argument('comment', metavar = 'COMMENT',
        action = 'store', type = str, nargs = '?', default = None,
        help = 'Optional comment.')

    return parser

if (__name__ == '__main__'):
    sys.exit(main())



import sys

import canvas.api.gradebook.fetch
import canvas.config

DEFAULT_SKIP_HEADERS = False
DEFAULT_INCLUDE_EMPTY_ASSIGNMENTS = False
DEFAULT_INCLUDE_EMPTY_USERS = False

def run_cli(include_empty_assignments = DEFAULT_INCLUDE_EMPTY_ASSIGNMENTS, include_empty_users = DEFAULT_INCLUDE_EMPTY_USERS,
        skip_headers = DEFAULT_SKIP_HEADERS, students = [], **kwargs):
    assignments, user_grades = canvas.api.gradebook.fetch.request(user_queries = students, **kwargs)

    if (len(assignments) == 0):
        print("No assignments found.", file = sys.stderr)
        return 0

    if (len(user_grades) == 0):
        print("No submissions found.", file = sys.stderr)
        return 0

    sorted_assignments = list(sorted(assignments.values(), key = lambda assignment: (assignment['assignment_group_id'], assignment['position'], assignment['id'])))

    if (not include_empty_assignments):
        new_assignments = []

        for assignment in sorted_assignments:
            if (assignment['count'] > 0):
                new_assignments.append(assignment)

        sorted_assignments = new_assignments

    if (not skip_headers):
        headers = ["%s (%s)" % (assignment['name'], str(assignment['id'])) for assignment in sorted_assignments]
        print("\t".join(['email'] + headers))

    for (email, grades) in user_grades.items():
        row = []
        has_score = False

        for assignment in sorted_assignments:
            score = grades.get(assignment['id'], None)
            if (score is None):
                score = ''
            else:
                has_score = True

            row.append(str(score))

        if ((not include_empty_users) and (not has_score)):
            continue

        print("\t".join([email] + row))

    return 0

def main():
    config = canvas.config.get_config(exit_on_error = True, modify_parser = _modify_parser, course = True)
    return run_cli(**config)

def _modify_parser(parser):
    parser.description = 'Fetch the gradebook for a course. This may take several minutes.'

    parser.add_argument('--include-empty-assignments', dest = 'include_empty_assignments',
        action = 'store_true', default = DEFAULT_INCLUDE_EMPTY_ASSIGNMENTS,
        help = 'Skip assignments with no submissions (default: %(default)s).')

    parser.add_argument('--include-empty-users', dest = 'include_empty_users',
        action = 'store_true', default = DEFAULT_INCLUDE_EMPTY_USERS,
        help = 'Skip users with no submissions (default: %(default)s).')

    parser.add_argument('--skip-headers', dest = 'skip_headers',
        action = 'store_true', default = DEFAULT_SKIP_HEADERS,
        help = 'Skip headers (default: %(default)s).')

    parser.add_argument('students',
        nargs = '*',
        help = 'If specified, only fetch the gradebook for the specified students.')

    return parser

if (__name__ == '__main__'):
    sys.exit(main())



import logging

import canvas.api.common
import canvas.api.user.common

BASE_ENDPOINT = "/api/v1/courses/{course}/users?per_page={page_size}"

def request(server = None, token = None, course = None,
        include_role = False,
        keys = canvas.api.user.common.DEFAULT_KEYS, **kwargs):
    server = canvas.api.common.validate_param(server, 'server')
    token = canvas.api.common.validate_param(token, 'token')
    course = canvas.api.common.validate_param(course, 'course', param_type = int)

    logging.info("Fetching course ('%s') users from '%s'." % (str(course), server))

    url = server + BASE_ENDPOINT.format(course = course, page_size = canvas.api.common.DEFAULT_PAGE_SIZE)
    headers = canvas.api.common.standard_headers(token)

    if (include_role):
        url += '&include[]=enrollments'

    return canvas.api.user.common.list_users(url, headers, keys)



import argparse
import logging
import os
import sys
import urllib.parse

import json5
import platformdirs

import canvas.log

DEFAULT_CONFIG_FILENAME = 'edq-canvas.json'
DEFAULT_USER_CONFIG_PATH = platformdirs.user_config_dir(DEFAULT_CONFIG_FILENAME)

CONFIG_PATHS_KEY = 'config_paths'

REQUIRED_KEYS = ['server', 'token']

def get_config(exit_on_error = False, modify_parser = None, **parser_options):
    """
    Collect config values from files and the command line.
    """

    try:
        return _get_config(modify_parser = modify_parser, **parser_options)
    except Exception as ex:
        if (exit_on_error):
            print("ERROR: " + ex.args[0], file = sys.stderr)
            sys.exit(1)

        raise ValueError("Failed to get config") from ex

def _get_config(modify_parser = None, **parser_options):
    parser = get_argument_parser(**parser_options)

    if (modify_parser is not None):
        modify_parser(parser)

    arguments = parser.parse_args()
    config = get_tiered_config(arguments)

    # Work with and clean standard options,
    canvas.log.init(level = config.get('log_level', canvas.log.DEFAULT_LOGGING_LEVEL))

    for key in REQUIRED_KEYS:
        if (key not in config):
            raise ValueError("Missing required config value: '%s'." % (key))

    try:
        parts = urllib.parse.urlparse(config['server'], scheme = 'https')
        if (parts.netloc != ''):
            config['server'] = parts.geturl()
        else:
            config['server'] = "%s://%s" % (parts.scheme, parts.path)
    except Exception as ex:
        raise ValueError("Failed to parse Canvas server: '%s'." % config['server']) from ex

    return config

def get_tiered_config(cli_arguments, skip_keys = [CONFIG_PATHS_KEY]):
    """
    Get all the tiered configuration options (from files and CLI).
    """

    config = {}

    if (isinstance(cli_arguments, argparse.Namespace)):
        cli_arguments = vars(cli_arguments)

    # Check the current directory config.
    if (os.path.isfile(DEFAULT_CONFIG_FILENAME)):
        with open(DEFAULT_CONFIG_FILENAME, 'r') as file:
            for key, value in json5.load(file).items():
                config[key] = value

    # Check the user config file.
    if (os.path.isfile(DEFAULT_USER_CONFIG_PATH)):
        with open(DEFAULT_USER_CONFIG_PATH, 'r') as file:
            for key, value in json5.load(file).items():
                config[key] = value

    # Check the config files specified on the command-line.
    config_paths = cli_arguments.get(CONFIG_PATHS_KEY, [])
    if (config_paths is not None):
        for path in config_paths:
            with open(path, 'r') as file:
                for key, value in json5.load(file).items():
                    config[key] = value

    # Finally, any command-line options.
    for (key, value) in cli_arguments.items():
        if (key in skip_keys):
            continue

        if ((value is None) or (value == '')):
            continue

        config[key] = value

    return config

def get_argument_parser(description = '', course = True, **kwargs):
    """
    Create an argparse parser that has all the standard options for API requests.
    """

    parser = argparse.ArgumentParser(description = description)

    parser.add_argument('--config', dest = CONFIG_PATHS_KEY,
        action = 'append', type = str,
        help = 'A JSON config file with your submission/authentication details.'
            + " Can be specified multiple times with later values overriding earlier ones."
            + " Config values can be specified in multiple places"
            + " (with later values overriding earlier values):"
            + " First './%s'," % (DEFAULT_CONFIG_FILENAME)
            + " then '%s'," % (DEFAULT_USER_CONFIG_PATH)
            + " now any files specified using --config in the order they were specified,"
            + " and finally any variables specified directly on the command line (like --user).")

    parser.add_argument('--log-level', dest = 'log_level',
        action = 'store', type = str, default = canvas.log.DEFAULT_LOGGING_LEVEL,
        choices = canvas.log.LEVELS,
        help = 'The logging level (default: %(default)s).')

    parser.add_argument('--server', dest = 'server',
        action = 'store', type = str, default = None,
        help = 'The Canvas instance to work with.')

    parser.add_argument('--token', dest = 'token',
        action = 'store', type = str, default = None,
        help = 'A Canvas authentication token.')

    if (course):
        parser.add_argument('--course', dest = 'course',
            action = 'store', type = int, default = None,
            help = 'The Canvas course ID.')

    return parser



import logging
import re

import canvas.api.assignment.common
import canvas.api.assignment.list

def fetch_and_resolve_assignments(server, token, course, assignment_queries,
        keys = canvas.api.assignment.common.DEFAULT_KEYS):
    course_assignments = canvas.api.assignment.list.request(server = server, token = token,
            course = course, keys = None)

    assignments = resolve_assignments(assignment_queries, course_assignments)

    if (keys is None):
        return assignments

    return [{key: assignment[key] for key in keys} for assignment in assignments]

def resolve_assignments(assignment_queries, course_assignments, id_field = 'id'):
    """
    Try to match each assignment query to an acutal assignment.
    """

    assignments = []

    for query in assignment_queries:
        id = None
        name = None

        # Clean whitespace.
        query = re.sub(r'\s+', ' ', str(query)).strip()
        if (query == ''):
            continue

        match = re.search(r'^(\S.*)\((\d+)\)$', query)
        if (match is not None):
            name = match.group(1).strip()
            id = match.group(2)
        elif (re.search(r'^\d+$', query) is not None):
            # Query must be an ID.
            id = query
        else:
            # Otherwise, query must be a name.
            name = query

        found_assignments = {}
        for assignment in course_assignments:
            if (id_field not in assignment):
                raise ValueError("Could not find id field '%s' in assignment ('%s')." % (id_field, str(assignment.keys())))
            assignment_id = assignment[id_field]

            if ((id is not None) and (assignment.get('id', '') == id)):
                found_assignments[assignment_id] = assignment

            if (name is not None):
                names = [
                    assignment.get('name', ''),
                ]

                if (name in names):
                    found_assignments[assignment_id] = assignment

        if (len(found_assignments) == 0):
            logging.warning("No assignment found that matches assignment query: '%s'." % (query))
        elif (len(found_assignments) > 1):
            labels = ["%s (%s)" % (str(assignment.get('name', '')), str(assignment.get('id', ''))) for assignment in found_assignments.values()]
            logging.warning("Assignment query ('%s') matches multiple assignments: '%s'." % (query, labels))
        else:
            assignments.append(list(found_assignments.values())[0])

    return assignments

def requires_resolution(assignments):
    """
    Does this list of assignments require resolution?
    I.E., does the list only contain assignment IDs?
    """

    for assignment in assignments:
        if (re.search(r'^\s*\d+\s*$', str(assignment)) is None):
            return True

    return False



import canvas.api.common

# Default keys that will be returned with a user.
DEFAULT_KEYS = [
    'id',
    'email',
    'name',
    'enrollment',
    'sis_user_id',
]

def list_users(url, headers, keys, missing_value = ''):
    output = []

    while (url is not None):
        _, url, new_users = canvas.api.common.make_get_request(url, headers)

        for new_user in new_users:
            raw_enrollments = new_user.get('enrollments', [])
            enrollment_types = [raw_enrollment.get('role') for raw_enrollment in raw_enrollments]
            new_user['enrollment'] = canvas.api.common.get_max_enrollment_type(enrollment_types)

            if (keys is not None):
                new_user = {key: new_user.get(key, missing_value) for key in keys}

            output.append(new_user)

    return output



import ast
import sys

import canvas.api.gradebook.upload
import canvas.config

def run_cli(path = None, **kwargs):
    assignments, users, scores = _load_gradebook(path)

    score_count = canvas.api.gradebook.upload.request(
            assignments = assignments, users = users, scores = scores,
            **kwargs)

    print("Uploaded %d Scores" % (score_count))
    return 0

# Return: ([assignment, ...], [user, ...], [[score, ...], ...])
# The scores are indexed by user then assignment: [user_index][assignment_index].
# Scores will always have all cells represented, a None indicates no score to upload.
def _load_gradebook(path):
    assignments = []
    users = []
    scores = []

    with open(path, 'r') as file:
        lineno = 0
        for line in file:
            lineno += 1

            if (line.strip() == ""):
                continue

            parts = [part.strip() for part in line.split("\t")]

            # First row has an empty cell then assignments.
            if (len(assignments) == 0):
                if (len(parts) <= 1):
                    raise ValueError("Gradebook '%s' has no assignments." % (path))

                for assignment in parts[1:]:
                    if (assignment == ''):
                        raise ValueError("Gradebook '%s' has an empty assignment name." % (path))

                    if (assignment in assignments):
                        raise ValueError("Gradebook '%s' has a duplicate assignment: '%s'." % (path, assignment))

                    assignments.append(assignment)

                continue

            if (len(parts) != (len(assignments) + 1)):
                raise ValueError(
                    "Gradebook '%s' line %d has the incorrect number of values." % (path, lineno)
                    + " Expecting %d, found %d." % ((len(assignments) + 1), len(parts)))

            user = parts.pop(0)
            if (user == ''):
                raise ValueError("Gradebook '%s' line %d has an empty user." % (path, lineno))

            if (user in users):
                raise ValueError("Gradebook '%s' line %d has a duplicate user '%s'." % (path, lineno, user))

            users.append(user)

            user_scores = []
            for i in range(len(parts)):
                score = None
                if (parts[i] != ''):
                    try:
                        score = ast.literal_eval(parts[i])
                    except Exception as ex:
                        raise ValueError(
                            "Gradebook '%s' line %d column %d has a score that cannot be" % (path, lineno, i + 2)
                            + " converted to a number: '%s'." % (parts[i])) from ex

                user_scores.append(score)

            scores.append(user_scores)

    return assignments, users, scores

def main():
    config = canvas.config.get_config(exit_on_error = True, modify_parser = _modify_parser, course = True)
    return run_cli(**config)

def _modify_parser(parser):
    parser.description = 'Upload a gradebook.'

    parser.epilog = ('A gradebook file is a tab-separated file where'
            + ' the first column is user queries (the first cell is ignored),'
            + ' the first row is assignment queries (the first cell is ignored),'
            + ' and the remaining cells are the associated scores.'
            + ' Any number of users and assignments can be specified as long as they exist in the course.'
            + ' Empty cells will not be uploaded.')

    parser.add_argument('path', metavar = 'PATH',
        action = 'store', type = str,
        help = 'Path to a TSV gradebook file.')

    return parser

if (__name__ == '__main__'):
    sys.exit(main())



import sys

import canvas.api.user.fetch
import canvas.cli.common
import canvas.cli.user.common
import canvas.config

DEFAULT_INCLUDE_ROLE = False
DEFAULT_TABLE = False
DEFAULT_SKIP_HEADERS = False

def run_cli(user = None, include_role = DEFAULT_INCLUDE_ROLE,
        table = DEFAULT_TABLE, skip_headers = DEFAULT_SKIP_HEADERS,
        **kwargs):
    raw_users = []
    if (user is not None):
        raw_users.append(user)

    users = canvas.api.user.fetch.request(users = raw_users, include_role = include_role,
            **kwargs)

    keys = canvas.cli.user.common.OUTPUT_KEYS.copy()
    if (include_role):
        keys.append(canvas.cli.user.common.ENROLLMENT_KEY)

    return canvas.cli.common.cli_list(users, keys,
            table = table, skip_headers = skip_headers,
            collective_name = 'user', sort_key = 'email')

def main():
    config = canvas.config.get_config(exit_on_error = True, modify_parser = _modify_parser)
    return run_cli(**config)

def _modify_parser(parser):
    parser.description = 'Fetch information for a user.'

    parser.add_argument('--include-role', dest = 'include_role',
        action = 'store_true', default = DEFAULT_INCLUDE_ROLE,
        help = 'Include user\'s role in the course (default: %(default)s).')

    parser.add_argument('-t', '--table', dest = 'table',
        action = 'store_true', default = DEFAULT_TABLE,
        help = 'Output the results as a TSV table with a header (default: %(default)s).')

    parser.add_argument('--skip-headers', dest = 'skip_headers',
        action = 'store_true', default = DEFAULT_SKIP_HEADERS,
        help = 'Skip headers when outputting as a table (default: %(default)s).')

    parser.add_argument('user',
        action = 'store', type = str,
        help = 'The query for the user to fetch information about.')

    return parser

if (__name__ == '__main__'):
    sys.exit(main())



import logging

import canvas.api.assignment.list
import canvas.api.common
import canvas.api.user.list
import canvas.api.user.resolve

BASE_ENDPOINT = "/api/v1/courses/{course}/students/submissions?per_page={page_size}"

# Get the current grades for all users/assignments.
# A count of all submissions will be returned for each assignment.
# Users will have a None for unsubmitted assignments.
# Return: (
#   {assignment_id: {id: <id>, name: <name>, assignment_group_id: <id>, count: <int>, position: <int>}, ...},
#   {user_email: {assignment_id: score, ...}, ...},
# )
def request(server = None, token = None, course = None, user_queries = [], **kwargs):
    server = canvas.api.common.validate_param(server, 'server')
    token = canvas.api.common.validate_param(token, 'token')
    course = canvas.api.common.validate_param(course, 'course', param_type = int)

    users = _fetch_users(server, token, course, user_queries)
    assignments = _fetch_assignments(server, token, course)

    logging.info("Fetching gradebook for course '%s' from '%s'." % (str(course), server))

    url = server + BASE_ENDPOINT.format(course = course, page_size = canvas.api.common.DEFAULT_PAGE_SIZE)
    headers = canvas.api.common.standard_headers(token)

    if (len(user_queries) == 0):
        url += "&student_ids[]=all"
    else:
        for user in users.values():
            url += "&student_ids[]=%s" % (user['id'])

    grades = {}

    while (url is not None):
        _, url, items = canvas.api.common.make_get_request(url, headers)

        for item in items:
            user_id = item.get('user_id', None)
            if (user_id is None):
                continue

            if (user_id not in users):
                continue

            user_email = users[user_id]['email']

            assignment_id = item.get('assignment_id', None)
            if (assignment_id is None):
                continue

            if (assignment_id not in assignments):
                continue

            assignment_name = assignments[assignment_id]['name']

            score = item.get('score', None)
            if (score is not None):
                score = float(score)

            if (user_email not in grades):
                grades[user_email] = {}

            if (score is not None):
                assignments[assignment_id]['count'] += 1

            grades[user_email][assignment_id] = score

    return assignments, grades

def _fetch_users(server, token, course, user_queries):
    if (len(user_queries) == 0):
        users = canvas.api.user.list.request(server = server, token = token, course = course)
    else:
        users = canvas.api.user.resolve.fetch_and_resolve_users(server, token, course, user_queries)

    if (len(users) == 0):
        raise ValueError("Could not find any users for the gradebook.")

    return {user['id']: user for user in users}

def _fetch_assignments(server, token, course):
    assignments = canvas.api.assignment.list.request(server = server, token = token, course = course)
    for assignment in assignments:
        assignment['count'] = 0

    return {assignment['id']: assignment for assignment in assignments}



import logging

import canvas.api.common
import canvas.api.user.common
import canvas.api.user.resolve

BASE_ENDPOINT = "/api/v1/courses/{course}/users?per_page={page_size}"

def request(server = None, token = None, course = None,
        users = [],
        include_role = False,
        keys = canvas.api.user.common.DEFAULT_KEYS, **kwargs):
    server = canvas.api.common.validate_param(server, 'server')
    token = canvas.api.common.validate_param(token, 'token')
    course = canvas.api.common.validate_param(course, 'course', param_type = int)

    logging.info("Fetching course ('%s') users (%s) from '%s'." % (
            str(course), ", ".join(map(str, users)), server))

    if (len(users) == 0):
        return []

    if (canvas.api.user.resolve.requires_resolution(users)):
        return canvas.api.user.resolve.fetch_and_resolve_users(
                server, token, course, users,
                keys = keys, include_role = include_role)

    url = server + BASE_ENDPOINT.format(course = course, page_size = canvas.api.common.DEFAULT_PAGE_SIZE)
    headers = canvas.api.common.standard_headers(token)

    if (include_role):
        url += '&include[]=enrollments'

    for user in users:
        url += "&user_ids[]=%s" % (str(user))

    return canvas.api.user.common.list_users(url, headers, keys)



import sys

# keys: [(items key, title, pretty title), ...]
def cli_list(items, keys, table = False, skip_headers = False,
        collective_name = 'items', sort_key = 'id'):
    if (len(items) == 0):
        print("No %s found." % (collective_name), file = sys.stderr)
        return 0

    if (table and (not skip_headers)):
        print("\t".join([key_set[1] for key_set in keys]))

    items = list(sorted(items, key = lambda item: item.get(sort_key, '')))
    for i in range(len(items)):
        item = items[i]

        values = {}
        for items_key, _, _ in keys:
            value = item[items_key]
            if (value is None):
                value = ''

            values[items_key] = str(value).strip()

        if (table):
            print("\t".join(map(_clean_cell, [values[key_set[0]] for key_set in keys])))
        else:
            if (i != 0):
                print()

            for key_set in keys:
                print("%s: %s" % (key_set[2], values[key_set[0]]))

    return 0

def _clean_cell(text):
    """
    Clean the text that will be in a TSV cell.
    """

    return str(text).replace("\t", " ").replace("\n", " ")



import sys

import canvas.api.assignment.list
import canvas.cli.assignment.common
import canvas.cli.common
import canvas.config

DEFAULT_TABLE = False
DEFAULT_SKIP_HEADERS = False
DEFAULT_SKIP_DESCRIPTION = False

def run_cli(table = DEFAULT_TABLE, skip_headers = DEFAULT_SKIP_HEADERS,
        skip_description = DEFAULT_SKIP_DESCRIPTION, **kwargs):
    assignments = canvas.api.assignment.list.request(**kwargs)

    keys = canvas.cli.assignment.common.OUTPUT_KEYS.copy()
    if (skip_description):
        keys = keys[:-1]

    return canvas.cli.common.cli_list(assignments, keys,
            table = table, skip_headers = skip_headers,
            collective_name = 'assignments', sort_key = 'name')

def main():
    config = canvas.config.get_config(exit_on_error = True, modify_parser = _modify_parser, course = True)
    return run_cli(**config)

def _modify_parser(parser):
    parser.description = 'List assignments in a course.'

    parser.add_argument('-t', '--table', dest = 'table',
        action = 'store_true', default = DEFAULT_TABLE,
        help = 'Output the results as a TSV table with a header (default: %(default)s).')

    parser.add_argument('--skip-headers', dest = 'skip_headers',
        action = 'store_true', default = DEFAULT_SKIP_HEADERS,
        help = 'Skip headers when outputting as a table (default: %(default)s).')

    parser.add_argument('--skip-description', dest = 'skip_description',
        action = 'store_true', default = DEFAULT_SKIP_DESCRIPTION,
        help = 'Skip outputting the assignment description (default: %(default)s).')

    return parser

if (__name__ == '__main__'):
    sys.exit(main())



import logging
import re

import canvas.api.user.common
import canvas.api.user.list

def fetch_and_resolve_users(server, token, course, user_queries,
        keys = canvas.api.user.common.DEFAULT_KEYS, include_role = False):
    course_users = canvas.api.user.list.request(server = server, token = token,
            course = course, include_role = include_role, keys = None)

    users = resolve_users(user_queries, course_users)

    if (keys is None):
        return users

    return [{key: user[key] for key in keys} for user in users]

def resolve_users(user_queries, course_users, id_field = 'id'):
    """
    Try to match each user query to an acutal user.
    """

    users = []

    for query in user_queries:
        id = None
        email = None
        name = None

        # Clean whitespace.
        query = re.sub(r'\s+', ' ', str(query)).strip()
        if (query == ''):
            continue

        match = re.search(r'^(\S.*)\((\d+)\)$', query)
        if (match is not None):
            email = match.group(1).strip()
            id = match.group(2)
        elif (re.search(r'^\d+$', query) is not None):
            # Query must be an ID.
            id = query
        elif ('@' in query):
            # Query should be an email.
            email = query
        else:
            # Otherwise, query must be a name.
            name = query

        found_users = {}
        for user in course_users:
            if (id_field not in user):
                raise ValueError("Could not find id field '%s' in user ('%s')." % (id_field, str(user.keys())))
            user_id = user[id_field]

            if ((id is not None) and (user.get('id', '') == id)):
                found_users[user_id] = user

            if (email is not None):
                emails = [
                    user.get('email', ''),
                    user.get('login_id', ''),
                ]

                if (email in emails):
                    found_users[user_id] = user

            if (name is not None):
                names = [
                    user.get('name', ''),
                    user.get('sortable_name', ''),
                    user.get('short_name', ''),
                ]

                if (name in names):
                    found_users[user_id] = user

        if (len(found_users) == 0):
            logging.warning("No user found that matches user query: '%s'." % (query))
        elif (len(found_users) > 1):
            labels = ["%s (%s)" % (str(user.get('email', '')), str(user.get('id', ''))) for user in found_users.values()]
            logging.warning("User query ('%s') matches multiple users: '%s'." % (query, labels))
        else:
            users.append(list(found_users.values())[0])

    return users

def requires_resolution(users):
    """
    Does this list of users require resolution?
    I.E., does the list only contain user IDs?
    """

    for user in users:
        if (re.search(r'^\s*\d+\s*$', str(user)) is None):
            return True

    return False



import sys

def main():
    print("You have invoked the `canvas` package directly. You probably want the `canvas.cli` package instead.")
    return 0

if (__name__ == '__main__'):
    sys.exit(main())



import canvas.api.common

# Default keys that will be returned with an assignment.
DEFAULT_KEYS = [
    'id',
    'name',
    'points_possible',
    'published',
    'due_at',
    'assignment_group_id',
    'position',
    'description',
]

def list_assignments(url, headers, keys):
    output = []

    while (url is not None):
        _, url, new_assignments = canvas.api.common.make_get_request(url, headers)

        for new_assignment in new_assignments:
            if (keys is not None):
                new_assignment = {key: new_assignment[key] for key in keys}

            output.append(new_assignment)

    return output



import logging

import canvas.api.assignment.common
import canvas.api.assignment.resolve
import canvas.api.common

BASE_ENDPOINT = "/api/v1/courses/{course}/assignments?per_page={page_size}"

def request(server = None, token = None, course = None,
        assignments = [],
        keys = canvas.api.assignment.common.DEFAULT_KEYS, **kwargs):
    server = canvas.api.common.validate_param(server, 'server')
    token = canvas.api.common.validate_param(token, 'token')
    course = canvas.api.common.validate_param(course, 'course', param_type = int)

    logging.info("Fetching course ('%s') assignments ('%s') from '%s'." % (
            str(course), ', '.join(map(str, assignments)), server))

    if (len(assignments) == 0):
        return []

    if (canvas.api.assignment.resolve.requires_resolution(assignments)):
        return canvas.api.assignment.resolve.fetch_and_resolve_assignments(
                server, token, course, assignments,
                keys = keys)

    url = server + BASE_ENDPOINT.format(course = course, page_size = canvas.api.common.DEFAULT_PAGE_SIZE)
    headers = canvas.api.common.standard_headers(token)

    for assignment in assignments:
        url += "&assignment_ids[]=%s" % (str(assignment))

    return canvas.api.assignment.common.list_assignments(url, headers, keys)



import importlib.util
import uuid

def import_path(path, module_name = None):
    if (module_name is None):
        module_name = str(uuid.uuid4()).replace('-', '')

    spec = importlib.util.spec_from_file_location(module_name, path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    return module



import json
import logging

import requests

DEFAULT_PAGE_SIZE = 95
HEADER_LINK = 'Link'

# Standard enrollment types order by priority.
ENROLLMENT_TYPES = [
    'ObserverEnrollment',
    'Student Auditor',
    'StudentEnrollment',
    'Tutor',
    'TaEnrollment',
    'DesignerEnrollment',
    'TA - Site Manager',
    'TeacherEnrollment',
]

ENROLLMENT_TYPE_UNKNOWN = '<unknown>'

# Return: (response, next_url, body)
def make_get_request(url, headers, raise_for_status = True,
        fetch_next_url = True, json_body = True):
    logging.info("Making request: '%s'." % (url))
    response = requests.get(url, headers = headers)
    response.raise_for_status()

    next_url = None
    if (fetch_next_url):
        next_url = fetch_next_canvas_link(response.headers)

    body = None
    log_body = None
    if (json_body):
        body = response.json()
        if (logging.getLogger().level <= logging.DEBUG):
            log_body = json.dumps(body, indent = 4)
    else:
        body = response.text
        log_body = body

    logging.debug("Response:\n%s" % log_body)

    return response, next_url, body

# Return: (response, body)
def make_post(url, headers, data,
        raise_for_status = True, json_body = True):
    logging.info("Making request: '%s'." % (url))
    logging.debug("Data:\n%s" % json.dumps(data, indent = 4))

    response = requests.post(url, headers = headers, data = data)
    response.raise_for_status()

    body = None
    log_body = None
    if (json_body):
        body = response.json()
        if (logging.getLogger().level <= logging.DEBUG):
            log_body = json.dumps(body, indent = 4)
    else:
        body = response.text
        log_body = body

    logging.debug("Response:\n%s" % log_body)

    return response, body

def validate_param(value, name, param_type = str, optional = False, strip = True):
    if (value is None):
        if (optional):
            return None

        raise ValueError("Parameter '%s' is missing or null/None." % (name))

    if (param_type == str):
        value = str(value)
        if (strip):
            value = value.strip()

        return value

    if (not isinstance(value, param_type)):
        raise ValueError("Parameter '%s' has the incorrect type '%s', expected '%s'." % (name, type(value), param_type))

    return value

def standard_headers(token):
    return {
        "Authorization": "Bearer %s" % (token),
        "Accept": "application/json+canvas-string-ids",
    }

def fetch_next_canvas_link(headers):
    if (HEADER_LINK not in headers):
        return None

    links = headers[HEADER_LINK].split(',')
    for link in links:
        parts = link.split(';')
        if (len(parts) != 2):
            continue

        if (parts[1].strip() != 'rel="next"'):
            continue

        return parts[0].strip().strip('<>')

    return None

# Given multiple enrollment types, return the one with the most authority.
def get_max_enrollment_type(enrollment_types):
    max_type = ENROLLMENT_TYPE_UNKNOWN
    max_index = -2

    for enrollment_type in enrollment_types:
        if (enrollment_type not in ENROLLMENT_TYPES):
            logging.warning("Unknown enrollment type: '%s'." % (enrollment_type))
            index = -1
        else:
            index = ENROLLMENT_TYPES.index(enrollment_type)

        if (index > max_index):
            max_index = index
            max_type = enrollment_type

    return max_type



"""
The `canvas.cli` package contains several packages for interacting with Canvas in differnt ways.
Each package can be invoked to list the tools (or subpackages) it contains.
Each tool includes a help prompt that can be accessed with the `-h`/`--help` flag.
"""

import sys

import canvas.util.cli

def main():
    canvas.util.cli.auto_list()
    return 0

if (__name__ == '__main__'):
    sys.exit(main())



# [(canvas key, title, pretty title), ...]
OUTPUT_KEYS = [
    ('email', 'email', 'Email'),
    ('name', 'name', 'Name'),
    ('id', 'canvas_id', 'Canvas ID'),
    ('sis_user_id', 'student_id', 'Student ID'),
]

ENROLLMENT_KEY = ('enrollment', 'role', 'Course Role')



import sys

import canvas.api.assignment.fetchscores
import canvas.config

DEFAULT_SKIP_HEADERS = False

def run_cli(skip_headers = DEFAULT_SKIP_HEADERS, **kwargs):
    submissions = canvas.api.assignment.fetchscores.request(**kwargs)

    if (len(submissions) == 0):
        print("No submissions found.", file = sys.stderr)
        return 0

    if (not skip_headers):
        header = "%s (%s)" % (submissions[0]['assignment']['name'], str(submissions[0]['assignment']['id']))
        print("\t".join(['email', header]))

    submissions = list(sorted(submissions, key = lambda submission: submission['user']['email']))
    for submission in submissions:
        score = submission.get('score', '')
        if (score is None):
            score = ''

        print("\t".join(map(str, [submission['user']['email'], score])))

    return 0

def main():
    config = canvas.config.get_config(exit_on_error = True, modify_parser = _modify_parser, course = True)
    return run_cli(**config)

def _modify_parser(parser):
    parser.description = 'List scores for an assignment.'

    parser.add_argument('--skip-headers', dest = 'skip_headers',
        action = 'store_true', default = DEFAULT_SKIP_HEADERS,
        help = 'Skip headers (default: %(default)s).')

    parser.add_argument('assignment',
        action = 'store', type = str,
        help = 'The query for the assignment to fetch information about.')

    return parser

if (__name__ == '__main__'):
    sys.exit(main())



import logging

import canvas.api.assignment.common
import canvas.api.common

BASE_ENDPOINT = "/api/v1/courses/{course}/assignments?per_page={page_size}"

def request(server = None, token = None, course = None,
        keys = canvas.api.assignment.common.DEFAULT_KEYS, **kwargs):
    server = canvas.api.common.validate_param(server, 'server')
    token = canvas.api.common.validate_param(token, 'token')
    course = canvas.api.common.validate_param(course, 'course', param_type = int)

    logging.info("Fetching course ('%s') assignments from '%s'." % (str(course), server))

    url = server + BASE_ENDPOINT.format(course = course, page_size = canvas.api.common.DEFAULT_PAGE_SIZE)
    headers = canvas.api.common.standard_headers(token)

    return canvas.api.assignment.common.list_assignments(url, headers, keys)



import logging
import requests

import canvas.api.assignment.resolve
import canvas.api.assignment.uploadscores
import canvas.api.common
import canvas.api.user.resolve

# Scores are a full 2d list indexed by user then assignment, None entries indicate a missing score.
def request(server = None, token = None, course = None,
        assignments = [], users = [], scores = [],
        **kwargs):
    server = canvas.api.common.validate_param(server, 'server')
    token = canvas.api.common.validate_param(token, 'token')
    course = canvas.api.common.validate_param(course, 'course', param_type = int)

    _validate_shape(assignments, users, scores)

    original_users = users
    original_assignments = assignments

    logging.info("Uploading gradebook for course '%s' from '%s'." % (str(course), server))

    if (canvas.api.assignment.resolve.requires_resolution(assignments)):
        resolved_assignments = canvas.api.assignment.resolve.fetch_and_resolve_assignments(server, token, course, assignments)
        if (len(resolved_assignments) != len(assignments)):
            raise ValueError("Unable to resolve all assignment queries.")

        assignments = [assignment['id'] for assignment in resolved_assignments]

    if (canvas.api.user.resolve.requires_resolution(users)):
        resolved_users = canvas.api.user.resolve.fetch_and_resolve_users(server, token, course, users)
        if (len(resolved_users) != len(users)):
            raise ValueError("Unable to resolve all user queries.")

        users = [resolved_user['id'] for resolved_user in resolved_users]

    _validate_dups(users, original_users, 'users')
    _validate_dups(assignments, original_assignments, 'assignments')

    score_count = 0

    for assignment_index in range(len(assignments)):
        assignment = assignments[assignment_index]
        assignment_scores = [scores[user_index][assignment_index] for user_index in range(len(users))]

        score_count += canvas.api.assignment.uploadscores.direct_request(server, token, course,
                assignment, users, assignment_scores)

    return score_count

def _validate_shape(assignments, users, scores):
    if (len(scores) != len(users)):
        raise ValueError("Scores does not match the number of users (%d), found %d rows." % (len(users), len(scores)))

    for i in range(len(scores)):
        user = users[i]
        assignment_scores = scores[i]

        if (len(assignment_scores) != len(assignments)):
            raise ValueError("Scores for user '%s' does not match the number of assignment (%d), found %d scores." % (
                    str(user), len(assignments), len(assignment_scores)))

def _validate_dups(resolved, original, label):
    dup_indexes = {}

    for i in range(len(resolved)):
        if (resolved[i] not in dup_indexes):
            dup_indexes[resolved[i]] = []
        dup_indexes[resolved[i]].append(i)

    dups = []
    for indexes in dup_indexes.values():
        if (len(indexes) > 1):
            dups.append("(%s)" % (', '.join([str(original[index]) for index in indexes])))

    if (len(dups) == 0):
        return

    raise ValueError("Found duplicates in %s: [%s]." % (label, ', '.join(dups)))



import logging
import requests

import canvas.api.assignment.resolve
import canvas.api.common
import canvas.api.user.resolve

BASE_ENDPOINT = "/api/v1/courses/{course}/assignments/{assignment}/submissions/update_grades"

def request(server = None, token = None, course = None, assignment = None,
        users = [], scores = [], comments = [],
        **kwargs):
    server = canvas.api.common.validate_param(server, 'server')
    token = canvas.api.common.validate_param(token, 'token')
    course = canvas.api.common.validate_param(course, 'course', param_type = int)
    assignment = canvas.api.common.validate_param(assignment, 'assignment')

    logging.info("Uploading %d scores for assignment ('%s' (course '%s')) from '%s'." % (len(users), assignment, str(course), server))

    if (canvas.api.assignment.resolve.requires_resolution([assignment])):
        resolved_assignments = canvas.api.assignment.resolve.fetch_and_resolve_assignments(server, token, course, [assignment])
        if (len(resolved_assignments) == 0):
            raise ValueError("Unable to resolve assignment query '%s'." % (assignment))

        assignment = resolved_assignments[0]['id']

    if (canvas.api.user.resolve.requires_resolution(users)):
        resolved_users = canvas.api.user.resolve.fetch_and_resolve_users(server, token, course, users)
        if (len(resolved_users) == 0):
            raise ValueError("Unable to resolve all user queries.")

        users = [resolved_user['id'] for resolved_user in resolved_users]

    return direct_request(server, token, course, assignment, users, scores, comments = comments)

# Perform the request with no resolution and minimal validation.
def direct_request(server, token, course, assignment, users, scores, comments = None):
    if ((len(users) != len(scores)) or ((comments is not None) and (len(users) != len(comments)))):
        raise ValueError("Mismatched count of users (%d), scores (%d), and comments (%d)." % (len(users), len(scores), len(comments)))

    score_count = 0
    data = {}
    for i in range(len(users)):
        if ((users[i] is None) or (scores[i] is None)):
            continue

        data["grade_data[%s][posted_grade]" % (users[i])] = scores[i]
        score_count += 1

        if ((comments is None) or (comments[i] is None)):
            continue

        comment = comments[i].strip()
        if (comment == ''):
            continue

        data["grade_data[%s][text_comment]" % (users[i])] = comment

    if (score_count == 0):
        return 0

    url = server + BASE_ENDPOINT.format(course = course, assignment = assignment)
    headers = canvas.api.common.standard_headers(token)

    canvas.api.common.make_post(url, headers, data)

    return score_count



import logging
import requests

import canvas.api.assignment.common
import canvas.api.assignment.fetch
import canvas.api.common
import canvas.api.user.common
import canvas.api.user.list

BASE_ENDPOINT = "/api/v1/courses/{course}/assignments/{assignment}/submissions?per_page={page_size}"

DEFAULT_KEYS = [
    'id',
    'grade',
    'score',
    'submitted_at',
    'graded_at',
]

def request(server = None, token = None, course = None, assignment = None,
        keys = DEFAULT_KEYS,
        assignment_keys = canvas.api.assignment.common.DEFAULT_KEYS, user_keys = canvas.api.user.common.DEFAULT_KEYS,
        **kwargs):
    server = canvas.api.common.validate_param(server, 'server')
    token = canvas.api.common.validate_param(token, 'token')
    course = canvas.api.common.validate_param(course, 'course', param_type = int)
    assignment = canvas.api.common.validate_param(assignment, 'assignment')

    logging.info("Fetching scores for assignement ('%s' (course '%s')) from '%s'." % (assignment, str(course), server))

    resolved_assignments = canvas.api.assignment.fetch.request(server = server, token = token, course = course,
            assignments = [assignment], keys = assignment_keys)

    if (len(resolved_assignments) == 0):
        raise ValueError("Unable to resolve assignment query '%s'." % (assignment))

    assignment_info = resolved_assignments[0]

    all_users = _fetch_users(server, token, course)

    url = server + BASE_ENDPOINT.format(course = course, assignment = assignment_info['id'],
            page_size = canvas.api.common.DEFAULT_PAGE_SIZE)
    headers = canvas.api.common.standard_headers(token)

    submissions = []

    while (url is not None):
        _, url, items = canvas.api.common.make_get_request(url, headers)

        for item in items:
            user_id = item.get('user_id', None)
            if (user_id is None):
                continue

            if (user_id not in all_users):
                continue

            item = {key: item[key] for key in keys}
            user = {key: all_users[user_id][key] for key in user_keys}

            item['user'] = user
            item['assignment'] = assignment_info

            submissions.append(item)

    return submissions

def _fetch_users(server, token, course):
    users = canvas.api.user.list.request(server = server, token = token, course = course)
    return {user['id']: user for user in users}



"""
Look in a package for modules that look like CLI tools and list their information.
A package looks like a CLI package if it has a __main__.py file.
A module looks like a CLI tool if it either
has a _get_parser() method that returns an argparse parser,
or has a _modify_parser() method that takes a copy of the default (passed in) argparse parser.
"""

import argparse
import copy
import importlib.util
import inspect
import os
import uuid

def list_dir(base_dir, command_prefix, default_parser):
    for dirent in sorted(os.listdir(base_dir)):
        path = os.path.join(base_dir, dirent)
        cmd = command_prefix + '.' + os.path.splitext(dirent)[0]

        if (dirent.startswith('__')):
            continue

        if (os.path.isfile(path)):
            _handle_file(path, cmd, default_parser)
        else:
            _handle_dir(path, cmd)

def auto_list(default_parser = None):
    """
    Will print the caller's prompt and call list_dir() on it,
    but will figure out the package's prompt (doc string), base_dir, and command_prefix automatially.
    This will use the inspect library, so only use in places that use code normally.
    """

    try:
        frameInfo = inspect.stack()[1]

        path = frameInfo.filename
        base_dir = os.path.dirname(path)

        module = inspect.getmodule(frameInfo.frame)
        package = module.__package__
    except Exception as ex:
        raise ValueError("Unable to get caller information for listing CLI information.") from ex

    if (default_parser is None):
        default_parser = argparse.ArgumentParser()

    print(module.__doc__.strip())
    list_dir(base_dir, package, default_parser)

def _handle_file(path, cmd, default_parser):
    if (not path.endswith('.py')):
        return

    try:
        module = _import_path(path)
    except Exception:
        print("ERROR Importing: ", path)
        return

    parser = None
    if ('_get_parser' in dir(module)):
        parser = module._get_parser()
    elif ('_modify_parser' in dir(module)):
        parser = copy.deepcopy(default_parser)
        module._modify_parser(parser)
    else:
        return

    parser.prog = 'python3 -m ' + cmd

    print()
    print(cmd)
    print(parser.description)
    parser.print_usage()

def _handle_dir(path, cmd):
    try:
        module = _import_path(os.path.join(path, '__main__.py'))
    except Exception:
        return

    description = module.__doc__.strip()

    print()
    print(cmd + '.*')
    print(description)
    print("See `python3 -m %s` for more information." % (cmd))

def _import_path(path, module_name = None):
    if (module_name is None):
        module_name = str(uuid.uuid4()).replace('-', '')

    spec = importlib.util.spec_from_file_location(module_name, path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    return module



import logging

DEFAULT_LOGGING_LEVEL = logging.getLevelName(logging.WARNING)
DEFAULT_LOGGING_FORMAT = '%(asctime)s [%(levelname)-8s] - %(filename)s:%(lineno)s -- %(message)s'

LEVELS = [
    logging.getLevelName(logging.DEBUG),
    logging.getLevelName(logging.INFO),
    logging.getLevelName(logging.WARNING),
    logging.getLevelName(logging.ERROR),
    logging.getLevelName(logging.CRITICAL),
]

def init(level = DEFAULT_LOGGING_LEVEL, format = DEFAULT_LOGGING_FORMAT, **kwargs):
    """
    Initialize or re-initialize the logging infrastructure.
    """

    logging.basicConfig(level = level, format = format, force = True)

    # Ignore logging from third-party libraries.
    logging.getLogger("urllib3").setLevel(logging.WARNING)

# Load the default logging when this module is loaded.
init()
